import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';
import { GUI } from './jsm/libs/dat.gui.module.js';

class App {
    constructor() {
        const domWebGL = document.createElement('div');
        document.body.appendChild(domWebGL);
        this._domWebGL = domWebGL;

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        //renderer.outputEncoding = THREE.sRGBEncoding;
        renderer.setPixelRatio(window.devicePixelRatio);
        //renderer.shadowMap.enabled =  true;
        domWebGL.appendChild(renderer.domElement);

        if(!renderer.capabilities.isWebGL2 && !renderer.extenstions.has("ANGLE_instanced_arrays")) {
            alert("Not Supported");
            return;
        }

        this._renderer = renderer;

        const scene = new THREE.Scene();
        //scene.background = new THREE.Color(0x050505);
        this._scene = scene;

        this._setupCamera();
        //this._setupLight();
        this._setupModel();
        //this._setupInteractions();
        this._setupGUI();
        //this._setupControls();
        //this._setupHelper();
        
        window.onresize = this.resize.bind(this);

        this.resize();
        this.render()
    }

    _setupInteractions() {
        //.
    }

    _setupPostprocessing() {
        //.
    }

    _setupModel() {
        const vector = new THREE.Vector4();
        const instances = 50000;
        const positions = [];
        const offsets = [];
        const colors = [];
        const orientationStart = [];
        const orientationEnd = [];

        positions.push(0.025, -0.025, 0);
        positions.push(-0.025, 0.025, 0);
        positions.push(0, 0, 0.025);

        for(let i=0; i<instances; i++) {
            offsets.push(Math.random()-0.5, Math.random()-0.5, Math.random()-0.5);
            
            colors.push(Math.random(), Math.random(), Math.random(), Math.random());
            
            vector.set(Math.random()*2-1, Math.random()*2-1, Math.random()*2-1, Math.random()*2-1);
            vector.normalize();
            orientationStart.push(vector.x, vector.y, vector.z, vector.w);

            vector.set(Math.random()*2-1, Math.random()*2-1, Math.random()*2-1, Math.random()*2-1);
            vector.normalize();
            orientationEnd.push(vector.x, vector.y, vector.z, vector.w);
        }

        const geometry = new THREE.InstancedBufferGeometry();
        geometry.instanceCount = instances;
        
        geometry.setAttribute("position", new THREE.Float32BufferAttribute(new Float32Array(positions), 3));
        geometry.setAttribute("offset", new THREE.InstancedBufferAttribute(new Float32Array(offsets), 3));
        geometry.setAttribute("color", new THREE.InstancedBufferAttribute( new Float32Array(colors), 4 ) );
        geometry.setAttribute("orientationStart", new THREE.InstancedBufferAttribute(new Float32Array(orientationStart), 4));
        geometry.setAttribute("orientationEnd", new THREE.InstancedBufferAttribute(new Float32Array(orientationEnd), 4));

        const vertexShader = `
            precision highp float;

            uniform float sineTime;

            uniform mat4 modelViewMatrix;
            uniform mat4 projectionMatrix;

            attribute vec3 position;
            attribute vec3 offset;
            attribute vec4 color;
            attribute vec4 orientationStart;
            attribute vec4 orientationEnd;

            varying vec3 vPosition;
            varying vec4 vColor;

            void main(){
                vPosition = offset * max(abs(sineTime), 0.1) + position;
                vec4 orientation = normalize( mix( orientationStart, orientationEnd, sineTime ) );
                vec3 vcV = cross( orientation.xyz, vPosition );
                vPosition = vcV * ( 2.0 * orientation.w ) + ( cross( orientation.xyz, vcV ) * 2.0 + vPosition );

                vColor = color;

                gl_Position = projectionMatrix * modelViewMatrix * vec4( vPosition, 1.0 );
            }
        `;

        const fragmentShader = `
            precision highp float;

            uniform float time;

            varying vec3 vPosition;
            varying vec4 vColor;

            void main() {
                vec4 color = vec4( vColor );
                color.r += sin( vPosition.x * 10.0 + time ) * 0.9;

                gl_FragColor = color;
            }
        `;

        const material = new THREE.RawShaderMaterial({
            uniforms: {
                "time": { value: 1.0 },
                "sineTime": { value: 1.0 }
            },
            vertexShader: vertexShader,
            fragmentShader: fragmentShader,
            side: THREE.DoubleSide,
            transparent: true
        });

        const mesh = new THREE.Mesh(geometry, material);
        this._scene.add(mesh);

        this._geometry = geometry;
        this._instances = instances;
    }

    _setupHelper() {
        //.
    }

    _setupGUI() {
        const stats = new Stats();
        this._domWebGL.appendChild(stats.dom);
        this._stats = stats;

        const gui = new GUI({ width: 350 });
        gui.add(this._geometry, "instanceCount", 0, this._instances);
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(50, window.innerWidth / window.innerHeight, 1, 10);
        camera.position.z = 2;

        this._camera = camera;
    }

    _setupControls() {
        //.
    }

    _setupLight() {
        //.
    }

    update() {
        const time = performance.now();
        const object = this._scene.children[0];

        object.rotation.y = time * 0.0005;
        object.material.uniforms["time"].value = time * 0.005;
        object.material.uniforms["sineTime"].value = Math.sin(object.material.uniforms["time"].value * 0.05);

        this._stats.update();
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._camera);
    }

    resize() {
        this._camera.aspect = window.innerWidth / window.innerHeight;
        this._camera.updateProjectionMatrix();
        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }
}

window.onload = function () {
    new App();
}